from unittest import mock

from django.test import TestCase
from django.test.utils import override_settings

from ambient_toolbox.management.commands.install_permission_fixtures import Command
from ambient_toolbox.permissions.fixtures.services import PermissionSetupService


class InstallPermissionFixturesCommandTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

    @override_settings(GROUP_PERMISSION_FIXTURES=['testapp.permissions.TestGroupDeclaration'])
    @mock.patch.object(PermissionSetupService, 'process', return_value=([], []))
    def test_run_command_regular(self, mocked_process):
        command = Command()
        command.handle()

        mocked_process.assert_called_once()

    @mock.patch.object(PermissionSetupService, 'process')
    def test_run_command_no_settings_variable(self, mocked_process):
        command = Command()
        command.handle()

        mocked_process.assert_not_called()



from ambient_package_update.metadata.author import PackageAuthor
from ambient_package_update.metadata.constants import DEV_DEPENDENCIES
from ambient_package_update.metadata.package import PackageMetadata
from ambient_package_update.metadata.readme import ReadmeContent
from ambient_package_update.metadata.ruff_ignored_inspection import RuffIgnoredInspection

METADATA = PackageMetadata(
    package_name='ambient_toolbox',
    authors=[
        PackageAuthor(
            name='Ambient Digital',
            email='hello@ambient.digital',
        ),
    ],
    development_status='5 - Production/Stable',
    readme_content=ReadmeContent(
        tagline='Python toolbox of Ambient Digital containing an abundance of useful tools and gadgets.',
        content="""## Features

* Useful classes and mixins for Django admin
* Coverage script for GitLab
* Extensions for DRF and GraphQL
* Mailing backends for Django
* Management commands for validating a projects test structure
* Object ownership tracking with timestamping
* Pattern for improved workflow with Django ORM
* Helper and util functions for many different use-cases
* Sentry plugins
* django-scrubber wrapper class
* Mixins and test classes for django (class-based) views

# Migration from "ai_django_core"

This package was previously known as [ai_django_core](https://pypi.org/project/ai-django-core/). Due to the
misleading nature of the name, we chose to rename it with something more meaningful.

The migration is really simple, just:

* Install ambient-toolbox and remove the dependency to ai-django-core
* Search and replace all usages of `from ai_django_core...` to `from ambient_toolbox...`
* The class-based mail functionality was moved to a separate package called
[django-pony-express](https://pypi.org/project/django-pony-express/).
""",
    ),
    dependencies=[
        'Django>=2.2.28',
        'bleach>=1.4,<6',
        'python-dateutil>=2.5.3',
    ],
    optional_dependencies={
        'dev': [
            *DEV_DEPENDENCIES,
            'gevent~=22.10',
        ],
        'drf': [
            'djangorestframework>=3.8.2',
        ],
        'graphql': [
            'graphene-django>=2.2.0',
            'django-graphql-jwt>=0.2.1',
        ],
        'sentry': [
            'sentry-sdk>=1.19.1',
        ],
        'view-layer': [
            'django-crispy-forms>=1.4.0',
        ],
    },
    ruff_ignore_list=[
        RuffIgnoredInspection(key='N999', comment="Project name contains underscore, not fixable"),
        RuffIgnoredInspection(key='A003', comment="Django attributes shadow python builtins"),
        RuffIgnoredInspection(key='DJ001', comment="Django model text-based fields shouldn't be nullable"),
        RuffIgnoredInspection(key='B905', comment="Can be enabled when Python <=3.9 support is dropped"),
        RuffIgnoredInspection(key='DTZ001', comment="TODO will affect \"tz_today()\" method"),
        RuffIgnoredInspection(key='DTZ005', comment="TODO will affect \"tz_today()\" method"),
    ],
)



from ambient_toolbox.permissions.fixtures.declarations import GroupPermissionDeclaration, PermissionModelDeclaration


class TestGroupDeclaration(GroupPermissionDeclaration):
    name = ('group_1',)
    permission_list = [
        PermissionModelDeclaration(
            app_label='testapp',
            codename_list=['view_mysinglesignalmodel'],
            model='mysinglesignalmodel',
        ),
    ]



# Generated by Django 3.2.18 on 2023-04-27 12:45

import ambient_toolbox.mixins.models
import ambient_toolbox.mixins.validation
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='MyPermissionModelMixin',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'managed': False,
            },
            bases=(ambient_toolbox.mixins.models.PermissionModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ModelWithCleanMixin',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            bases=(ambient_toolbox.mixins.validation.CleanOnSaveMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ModelWithSaveWithoutSignalsMixin',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.PositiveIntegerField(default=0)),
            ],
            bases=(ambient_toolbox.mixins.models.SaveWithoutSignalsMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ModelWithSelector',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.PositiveIntegerField(default=0)),
            ],
        ),
        migrations.CreateModel(
            name='MyMultipleSignalModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.PositiveIntegerField(default=0)),
            ],
        ),
        migrations.CreateModel(
            name='MySingleSignalModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.PositiveIntegerField(default=0)),
            ],
        ),
        migrations.CreateModel(
            name='ModelWithOneToOneToSelf',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                (
                    'peer',
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='related_peer',
                        to='testapp.modelwithonetoonetoself',
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name='ModelWithFkToSelf',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                (
                    'parent',
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='children',
                        to='testapp.modelwithfktoself',
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name='ForeignKeyRelatedModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                (
                    'single_signal',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='foreign_key_related_models',
                        to='testapp.mysinglesignalmodel',
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name='CommonInfoBasedModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                (
                    'created_at',
                    models.DateTimeField(db_index=True, default=django.utils.timezone.now, verbose_name='Created at'),
                ),
                (
                    'lastmodified_at',
                    models.DateTimeField(
                        db_index=True, default=django.utils.timezone.now, verbose_name='Last modified at'
                    ),
                ),
                ('value', models.PositiveIntegerField(default=0)),
                (
                    'created_by',
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name='testapp_commoninfobasedmodel_created',
                        to=settings.AUTH_USER_MODEL,
                        verbose_name='Created by',
                    ),
                ),
                (
                    'lastmodified_by',
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name='testapp_commoninfobasedmodel_lastmodified',
                        to=settings.AUTH_USER_MODEL,
                        verbose_name='Last modified by',
                    ),
                ),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



SENTRY_EVENT = {
    'level': 'error',
    'exception': {
        'values': [
            {
                'module': None,
                'type': 'ZeroDivisionError',
                'value': 'division by zero',
                'mechanism': {'type': 'django', 'handled': False},
                'stacktrace': {
                    'frames': [
                        {
                            'filename': 'test/account/api/views.py',
                            'abs_path': '/opt/project/backend/test/account/api/views.py',
                            'function': 'perform_update',
                            'module': 'test.account.api.views',
                            'lineno': 123,
                            'pre_context': ['test = 1/0'],
                            'vars': {
                                'self': '<testapp.account.api.views.TestViewSet object>',
                                'serializer': 'TestDetailSerializer(<User: Test AAAA>)',
                                '__class__': "<class 'test.account.api.views.TestViewSet'>",
                                'admin': '<EmailUser: Test User>',
                                'test_data': '123',
                            },
                            'in_app': True,
                        }
                    ]
                },
            }
        ]
    },
    'request': {
        'url': 'http://localhost:8000/api/v2/test/17/',
        'query_string': '',
        'method': 'PUT',
        'env': {'SERVER_NAME': 'cdc3cb5b00af', 'SERVER_PORT': '8000', 'REMOTE_ADDR': '172.22.0.1'},
        'headers': {
            'Content-Length': '1202',
            'Content-Type': 'application/json',
            'Host': 'localhost:8000',
            'User-Agent': 'Mozilla/5.0 (Macintosh;) Gecko/20100101 Firefox/112.0',
            'Accept': 'application/json, text/plain, */*',
            'Accept-Language': 'en',
            'Accept-Encoding': 'gzip, deflate, br',
            'Authorization': 'Token 123123123123123123123123123123123123abcd',
            'Origin': 'http://127.0.0.1:3000',
            'Dnt': '1',
            'Connection': 'keep-alive',
            'Referer': 'http://127.0.0.1:3000/',
            'Sec-Fetch-Dest': 'empty',
            'Sec-Fetch-Mode': 'cors',
            'Sec-Fetch-Site': 'cross-site',
        },
        'cookies': {},
        'data': '',
    },
    'user': {'email': 'test@test.local', 'id': '1337', 'ip_address': '172.22.0.1', 'username': 'test@test.local'},
}

SCRUBBED_SENTRY_EVENT = {
    'level': 'error',
    'exception': {
        'values': [
            {
                'module': None,
                'type': 'ZeroDivisionError',
                'value': 'division by zero',
                'mechanism': {'type': 'django', 'handled': False},
                'stacktrace': {
                    'frames': [
                        {
                            'filename': 'test/account/api/views.py',
                            'abs_path': '/opt/project/backend/test/account/api/views.py',
                            'function': 'perform_update',
                            'module': 'test.account.api.views',
                            'lineno': 123,
                            'pre_context': ['test = 1/0'],
                            'vars': {
                                'self': "'<testapp.account.api.views.TestViewSet object>'",
                                'serializer': '[Filtered]',  # filtered due to standard filter
                                '__class__': '"<class \'test.account.api.views.TestViewSet\'>"',
                                'admin': '[Filtered]',  # filtered due to standard filter
                                'test_data': "'123'",
                            },
                            'in_app': True,
                        }
                    ]
                },
            }
        ]
    },
    'request': {
        'url': 'http://localhost:8000/api/v2/test/17/',
        'query_string': '',
        'method': 'PUT',
        'env': {'SERVER_NAME': 'cdc3cb5b00af', 'SERVER_PORT': '8000', 'REMOTE_ADDR': '172.22.0.1'},
        'headers': {
            'Content-Length': '1202',
            'Content-Type': 'application/json',
            'Host': 'localhost:8000',
            'User-Agent': 'Mozilla/5.0 (Macintosh;) Gecko/20100101 Firefox/112.0',
            'Accept': 'application/json, text/plain, */*',
            'Accept-Language': 'en',
            'Accept-Encoding': 'gzip, deflate, br',
            'Authorization': '[Filtered]',  # filtered due to default sentry filter
            'Origin': 'http://127.0.0.1:3000',
            'Dnt': '1',
            'Connection': 'keep-alive',
            'Referer': 'http://127.0.0.1:3000/',
            'Sec-Fetch-Dest': 'empty',
            'Sec-Fetch-Mode': 'cors',
            'Sec-Fetch-Site': 'cross-site',
        },
        'cookies': {},
        'data': '',
    },
    'user': {
        'email': '[Filtered]',  # filtered due to default sentry filter
        'id': '1337',
        'ip_address': '[Filtered]',  # filtered due to default sentry filter
        'username': '[Filtered]',  # filtered due to default sentry filter
    },
    '_meta': {
        'exception': {
            'values': {
                '0': {
                    'stacktrace': {
                        'frames': {
                            '0': {
                                'vars': {
                                    'serializer': {'': {'rem': [['!config', 's']]}},
                                    'admin': {'': {'rem': [['!config', 's']]}},
                                }
                            }
                        }
                    }
                }
            }
        },
        'request': {'headers': {'Authorization': {'': {'rem': [['!config', 's']]}}}},
        'user': {
            'email': {'': {'rem': [['!config', 's']]}},
            'ip_address': {'': {'rem': [['!config', 's']]}},
            'username': {'': {'rem': [['!config', 's']]}},
        },
    },
}



from typing import List


def generate_default_permissions(model_name: str) -> List[str]:
    return [
        f'add_{model_name}',
        f'change_{model_name}',
        f'delete_{model_name}',
        f'view_{model_name}',
    ]



from django.contrib.auth.models import AnonymousUser
from django.test import TestCase
from django.views import generic

from ambient_toolbox.tests.mixins import RequestProviderMixin
from ambient_toolbox.view_layer.htmx_mixins import HtmxResponseMixin


class HtmxResponseMixinTest(RequestProviderMixin, TestCase):
    class TestView(HtmxResponseMixin, generic.View):
        hx_redirect_url = 'https://my-url.com'
        hx_trigger = 'myEvent'

    class TestViewWithTriggerDict(HtmxResponseMixin, generic.View):
        hx_trigger = {'myEvent': None}

    def test_dispatch_functional(self):
        view = self.TestView()

        response = view.dispatch(request=self.get_request(user=AnonymousUser()))

        self.assertIn('HX-Redirect', response)
        self.assertEqual(response['HX-Redirect'], 'https://my-url.com')

        self.assertIn('HX-Trigger', response)
        self.assertEqual(response['HX-Trigger'], 'myEvent')

    def test_dispatch_trigger_with_dict(self):
        view = self.TestViewWithTriggerDict()

        response = view.dispatch(request=self.get_request(user=AnonymousUser()))

        self.assertIn('HX-Trigger', response)
        self.assertEqual(response['HX-Trigger'], "{\"myEvent\": null}")

    def test_get_hx_redirect_url_regular(self):
        view = self.TestView()

        self.assertEqual(view.get_hx_redirect_url(), 'https://my-url.com')

    def test_get_hx_trigger_regular(self):
        view = self.TestView()

        self.assertEqual(view.get_hx_trigger(), 'myEvent')



from pathlib import Path

from django.conf import settings
from django.test import TestCase, override_settings

from ambient_toolbox.tests.structure_validator.test_structure_validator import TestStructureValidator


class TestStructureValidatorTest(TestCase):
    def test_init_regular(self):
        service = TestStructureValidator()

        self.assertEqual(service.file_whitelist, ['__init__'])
        self.assertEqual(service.issue_list, [])

    @override_settings(TEST_STRUCTURE_VALIDATOR_FILE_WHITELIST=['my_file'])
    def test_get_file_whitelist_from_settings(self):
        service = TestStructureValidator()
        file_whitelist = service._get_file_whitelist()

        self.assertEqual(file_whitelist, ['__init__', 'my_file'])

    def test_get_file_whitelist_fallback(self):
        service = TestStructureValidator()
        file_whitelist = service._get_file_whitelist()

        self.assertEqual(file_whitelist, ['__init__'])

    @override_settings(TEST_STRUCTURE_VALIDATOR_BASE_DIR=settings.BASE_PATH)
    def test_get_base_dir_from_settings(self):
        service = TestStructureValidator()
        base_dir = service._get_base_dir()

        self.assertEqual(base_dir, settings.BASE_PATH)

    def test_get_base_dir_fallback(self):
        service = TestStructureValidator()
        base_dir = service._get_base_dir()

        self.assertEqual(base_dir, '')

    @override_settings(TEST_STRUCTURE_VALIDATOR_BASE_APP_NAME='my_project')
    def test_get_base_app_name_from_settings(self):
        service = TestStructureValidator()
        base_app_name = service._get_base_app_name()

        self.assertEqual(base_app_name, 'my_project')

    def test_get_base_app_name_fallback(self):
        service = TestStructureValidator()
        base_app_name = service._get_base_app_name()

        self.assertEqual(base_app_name, 'apps')

    @override_settings(TEST_STRUCTURE_VALIDATOR_APP_LIST=['apps.my_app', 'apps.other_app'])
    def test_get_app_list_from_settings(self):
        service = TestStructureValidator()
        base_app_name = service._get_app_list()

        self.assertEqual(base_app_name, ['apps.my_app', 'apps.other_app'])

    def test_get_app_list_fallback(self):
        service = TestStructureValidator()
        base_app_name = service._get_app_list()

        self.assertEqual(base_app_name, settings.INSTALLED_APPS)

    @override_settings(TEST_STRUCTURE_VALIDATOR_IGNORED_DIRECTORY_LIST=['my_dir', 'other_dir'])
    def test_get_ignored_directory_list_from_settings(self):
        service = TestStructureValidator()
        dir_list = service._get_ignored_directory_list()

        self.assertEqual(dir_list, ['__pycache__', 'my_dir', 'other_dir'])

    def test_get_ignored_directory_list_fallback(self):
        service = TestStructureValidator()
        dir_list = service._get_ignored_directory_list()

        self.assertEqual(dir_list, ['__pycache__'])

    def test_check_missing_test_prefix_correct_prefix(self):
        service = TestStructureValidator()
        result = service._check_missing_test_prefix(
            root='root/path',
            file='missing_prefix',
            filename='test_my_file',
            extension='.py',
        )

        self.assertTrue(result)
        self.assertEqual(len(service.issue_list), 0)

    @override_settings(TEST_STRUCTURE_VALIDATOR_FILE_WHITELIST=['my_file'])
    def test_check_missing_test_prefix_wrong_prefix_but_whitelisted(self):
        service = TestStructureValidator()
        result = service._check_missing_test_prefix(
            root='root/path',
            file='missing_prefix',
            filename='my_file',
            extension='.py',
        )

        self.assertTrue(result)
        self.assertEqual(len(service.issue_list), 0)

    def test_check_missing_test_prefix_wrong_prefix_but_not_py_file(self):
        service = TestStructureValidator()
        result = service._check_missing_test_prefix(
            root='root/path',
            file='missing_prefix',
            filename='missing_prefix',
            extension='.txt',
        )

        self.assertTrue(result)
        self.assertEqual(len(service.issue_list), 0)

    def test_check_missing_test_prefix_wrong_prefix(self):
        service = TestStructureValidator()
        result = service._check_missing_test_prefix(
            root='root/path',
            file='missing_prefix',
            filename='missing_prefix',
            extension='.py',
        )

        self.assertFalse(result)
        self.assertEqual(len(service.issue_list), 1)

    def test_check_missing_init_init_found_files_in_dir(self):
        service = TestStructureValidator()
        result = service._check_missing_init(root='root/path', init_found=True, number_of_py_files=1)

        self.assertTrue(result)
        self.assertEqual(len(service.issue_list), 0)

    def test_check_missing_init_no_init_no_files(self):
        service = TestStructureValidator()
        result = service._check_missing_init(root='root/path', init_found=False, number_of_py_files=0)

        self.assertTrue(result)
        self.assertEqual(len(service.issue_list), 0)

    def test_check_missing_init_no_init_but_files(self):
        service = TestStructureValidator()
        result = service._check_missing_init(root='root/path', init_found=False, number_of_py_files=1)

        self.assertFalse(result)
        self.assertEqual(len(service.issue_list), 1)

    @override_settings(
        TEST_STRUCTURE_VALIDATOR_BASE_DIR=Path('/src/ambient_toolbox/'),
        TEST_STRUCTURE_VALIDATOR_BASE_APP_NAME='my_project',
    )
    def test_build_path_to_test_package_with_settings_path(self):
        service = TestStructureValidator()
        path = service._build_path_to_test_package(app='my_project.my_app')

        self.assertEqual(path, Path('/src/ambient_toolbox/my_project/my_app/tests'))

    @override_settings(
        TEST_STRUCTURE_VALIDATOR_BASE_DIR='/src/ambient_toolbox/', TEST_STRUCTURE_VALIDATOR_BASE_APP_NAME='my_project'
    )
    def test_build_path_to_test_package_with_settings_str(self):
        service = TestStructureValidator()
        path = service._build_path_to_test_package(app='my_project.my_app')

        self.assertEqual(path, Path('/src/ambient_toolbox/my_project/my_app/tests'))

    def test_build_path_to_test_package_with_defaults(self):
        service = TestStructureValidator()
        path = service._build_path_to_test_package(app='my_project.my_app')

        self.assertEqual(path, Path('my_project/my_app/tests'))

    @override_settings(
        TEST_STRUCTURE_VALIDATOR_BASE_DIR=settings.BASE_PATH,
        TEST_STRUCTURE_VALIDATOR_APP_LIST=['testapp'],
        TEST_STRUCTURE_VALIDATOR_BASE_APP_NAME='',
    )
    def test_process_functional(self):
        service = TestStructureValidator()
        with self.assertRaises(SystemExit):
            service.process()

        self.assertEqual(len(service.issue_list), 2)

        complaint_list = sorted(service.issue_list)

        self.assertIn('Python file without "test_" prefix found:', complaint_list[0])
        self.assertIn('testapp/tests/subdirectory/missing_test_prefix.py', complaint_list[0])

        self.assertIn('__init__.py missing in', complaint_list[1])
        self.assertIn('testapp/tests/missing_init', complaint_list[1])



import dataclasses
from typing import List


@dataclasses.dataclass
class PermissionModelDeclaration:
    app_label: str
    codename_list: List[str]
    model: str


@dataclasses.dataclass
class GroupPermissionDeclaration:
    name: str
    permission_list: List[PermissionModelDeclaration]



from django import forms
from django.forms import BaseInlineFormSet, inlineformset_factory
from django.test import TestCase

from ambient_toolbox.view_layer.formset_mixins import CountChildrenFormsetMixin
from testapp.models import ForeignKeyRelatedModel, MySingleSignalModel


class ForeignKeyRelatedModelForm(forms.ModelForm):
    class Meta:
        model = ForeignKeyRelatedModel
        fields = ('single_signal',)


class MySingleSignalModelFormset(CountChildrenFormsetMixin, BaseInlineFormSet):
    pass


class CountChildrenFormsetMixinTest(TestCase):
    def test_simple_no_data(self):
        formset_class = inlineformset_factory(
            MySingleSignalModel,
            ForeignKeyRelatedModel,
            form=ForeignKeyRelatedModelForm,
            formset=MySingleSignalModelFormset,
            extra=3,
            can_delete=True,
            max_num=3,
        )

        formset = formset_class()
        self.assertEqual(formset.get_number_of_children(), 0)

    def test_regular_with_data(self):
        mssm = MySingleSignalModel.objects.create(value=27)
        ForeignKeyRelatedModel.objects.create(single_signal=mssm)
        ForeignKeyRelatedModel.objects.create(single_signal=mssm)

        formset_class = inlineformset_factory(
            MySingleSignalModel,
            ForeignKeyRelatedModel,
            form=ForeignKeyRelatedModelForm,
            formset=MySingleSignalModelFormset,
            extra=3,
            can_delete=True,
            max_num=3,
        )

        formset = formset_class(
            {'fkrm-INITIAL_FORMS': '2', 'fkrm-MIN_NUM_FORMS': '2', 'fkrm-MAX_NUM_FORMS': '3', 'fkrm-TOTAL_FORMS': '2'},
            None,
            instance=mssm,
            prefix='fkrm',
        )

        formset.is_valid()
        self.assertEqual(formset.get_number_of_children(), 2)



from django.core.mail import EmailMultiAlternatives
from django.db import models
from django.db.models.signals import post_save, pre_save
from django.dispatch import receiver

from ambient_toolbox.managers import GloballyVisibleQuerySet
from ambient_toolbox.mixins.models import PermissionModelMixin, SaveWithoutSignalsMixin
from ambient_toolbox.mixins.validation import CleanOnSaveMixin
from ambient_toolbox.models import CommonInfo
from testapp.managers import ModelWithSelectorQuerySet
from testapp.selectors import ModelWithSelectorGloballyVisibleSelector


class MySingleSignalModel(models.Model):
    value = models.PositiveIntegerField(default=0)

    objects = GloballyVisibleQuerySet.as_manager()

    def __str__(self):
        return str(self.value)


class ForeignKeyRelatedModel(models.Model):
    single_signal = models.ForeignKey(
        MySingleSignalModel, on_delete=models.CASCADE, related_name='foreign_key_related_models'
    )

    objects = GloballyVisibleQuerySet.as_manager()

    def __str__(self):
        return str(self.id)


@receiver(pre_save, sender=MySingleSignalModel)
def increase_value_no_dispatch_uid(sender, instance, **kwargs):
    instance.value += 1


class MyMultipleSignalModel(models.Model):
    value = models.PositiveIntegerField(default=0)

    def __str__(self):
        return str(self.value)


@receiver(pre_save, sender=MyMultipleSignalModel, dispatch_uid='test.mysinglesignalmodel.increase_value_with_uuid')
def increase_value_with_dispatch_uid(sender, instance, **kwargs):
    instance.value += 1


@receiver(post_save, sender=MyMultipleSignalModel)
def send_email(sender, instance, **kwargs):
    msg = EmailMultiAlternatives(
        'Test Mail', 'I am content', from_email='test@example.com', to=['random.dude@example.com']
    )
    msg.send()


class CommonInfoBasedModel(CommonInfo):
    value = models.PositiveIntegerField(default=0)

    def __str__(self):
        return str(self.value)


class ModelWithSelector(models.Model):
    value = models.PositiveIntegerField(default=0)

    objects = ModelWithSelectorQuerySet.as_manager()
    selectors = ModelWithSelectorGloballyVisibleSelector()

    def __str__(self):
        return str(self.value)


class ModelWithFkToSelf(models.Model):
    parent = models.ForeignKey('self', blank=True, null=True, related_name='children', on_delete=models.CASCADE)

    def __str__(self):
        return str(self.id)


class ModelWithOneToOneToSelf(models.Model):
    peer = models.OneToOneField('self', blank=True, null=True, related_name='related_peer', on_delete=models.CASCADE)

    def __str__(self):
        return str(self.id)


class ModelWithCleanMixin(CleanOnSaveMixin, models.Model):
    def __str__(self):
        return str(self.id)

    def clean(self):
        return True


class MyPermissionModelMixin(PermissionModelMixin, models.Model):
    pass

    def __str__(self):
        return str(self.id)


class ModelWithSaveWithoutSignalsMixin(SaveWithoutSignalsMixin, models.Model):
    value = models.PositiveIntegerField(default=0)

    def __str__(self):
        return str(self.value)


@receiver(pre_save, sender=ModelWithSaveWithoutSignalsMixin)
def increase_value_on_pre_save(sender, instance, **kwargs):
    instance.value += 1



from django.test import TestCase

from ambient_toolbox.permissions.fixtures.declarations import GroupPermissionDeclaration, PermissionModelDeclaration


class PermissionFixtureDeclarationTest(TestCase):
    def test_permission_model_declaration_regular(self):
        permission = PermissionModelDeclaration(
            app_label='my_app',
            codename_list=['view_mymodel'],
            model='mymodel',
        )

        self.assertEqual(permission.app_label, 'my_app')
        self.assertEqual(permission.codename_list, ['view_mymodel'])
        self.assertEqual(permission.model, 'mymodel')

    def test_group_permission_declaration_regular(self):
        permission = PermissionModelDeclaration(
            app_label='my_app',
            codename_list=['view_mymodel'],
            model='mymodel',
        )
        group = GroupPermissionDeclaration(
            name='my_group',
            permission_list=[permission],
        )

        self.assertEqual(group.name, 'my_group')
        self.assertEqual(group.permission_list, [permission])



import datetime
from unittest.mock import PropertyMock, patch

from django.test import TestCase
from freezegun import freeze_time

from testapp.models import CommonInfoBasedModel


class CommonInfoTest(TestCase):
    @freeze_time('2022-06-26 10:00')
    def test_save_update_fields_common_fields_set(self):
        with freeze_time('2020-09-19'):
            obj = CommonInfoBasedModel.objects.create(value=1)
        obj.value = 2
        obj.save(update_fields=('value',))

        obj.refresh_from_db()
        self.assertEqual(obj.value, 2)
        self.assertEqual(obj.lastmodified_at, datetime.datetime(2022, 6, 26, 10))

    @patch('testapp.models.CommonInfoBasedModel.ALWAYS_UPDATE_FIELDS', new_callable=PropertyMock)
    @freeze_time('2022-06-26 10:00')
    def test_save_update_fields_common_fields_set_without_always_update(self, always_update_mock):
        always_update_mock.return_value = False
        with freeze_time('2020-09-19'):
            obj = CommonInfoBasedModel.objects.create(value=1)
        obj.value = 2
        obj.save(update_fields=('value',))

        obj.refresh_from_db()
        self.assertEqual(obj.value, 2)
        self.assertEqual(obj.lastmodified_at, datetime.datetime(2020, 9, 19))

    @freeze_time('2022-06-26 10:00')
    def test_save_common_fields_set_without_update_fields(self):
        with freeze_time('2020-09-19'):
            obj = CommonInfoBasedModel.objects.create(value=1)
        obj.value = 2
        obj.save()

        obj.refresh_from_db()
        self.assertEqual(obj.value, 2)
        self.assertEqual(obj.lastmodified_at, datetime.datetime(2022, 6, 26, 10))
